# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# SSDF is distributed under the terms of the (new) BSD License.
# See http://www.opensource.org/licenses/bsd-license.php

""" ssdf.ssdf_base.py

Implements the base functionality for ssdf:
    * The Struct class
    * Some constants and small functions
    * The VirtualArray class
    * The base classes for SSDFReader, SSDFWriter anf Block
"""

import sys
from . import ClassManager

# Try importing numpy
try:
    import numpy as np
except ImportError:
    np = None

# From six.py
PY3 = sys.version_info[0] == 3
if PY3:
    string_types = str,
    integer_types = int,
    text_type = str
    binary_type = bytes
    ascii_type = str # Simple string
    from functools import reduce
else:
    string_types = basestring,  # noqa
    integer_types = (int, long)  # noqa
    text_type = unicode  # noqa
    binary_type = str
    ascii_type = str # Simple string
    reduce = reduce
    
# To store other classes
_CLASS_NAME = '_CLASS_NAME_'

# Same as in ssdf_bin (we only need the dict type id.
_TYPE_DICT = ord('D')

# Determine float and int types
if True:
    _FLOAT_TYPES = set([float])
    _INT_TYPES = set(integer_types)
if np:
    _FLOAT_TYPES.update([np.float32, np.float64])
    _INT_TYPES.update([   np.int8,  np.int16,  np.int32,  np.int64,
                        np.uint8, np.uint16, np.uint32, np.uint64 ])
_FLOAT_TYPES = tuple(_FLOAT_TYPES)
_INT_TYPES = tuple(_INT_TYPES)



def isstruct(ob):
    """ isstruct(ob)
    
    Returns whether the given object is an SSDF struct.
    
    """
    if hasattr(ob, '__is_ssdf_struct__'):
        return bool(ob.__is_ssdf_struct__)
    else:
        return False


def _not_equal(ob1, ob2):
    """ _not_equal(ob1, ob2)
    
    Returns None if the objects are equal. Otherwise returns a string
    indicating how the objects are inequal.
    
    """
    
    if isstruct(ob1) or isinstance(ob1, dict):
        if not ( isstruct(ob2) or isinstance(ob2, dict) ):
            return '<type does not match>'
        # Test number of elements
        keys1 = [key for key in ob1]
        keys2 = [key for key in ob2]
        if len(keys1) != len(keys2):
            return '<lengths do not match>'
        # Test all elements
        for key in keys1:
            if key not in keys2:
                return '<key not present in other struct/dict>'
            not_equal = _not_equal(ob1[key], ob2[key])
            if not_equal:
                return '.' + key + not_equal
    
    elif isinstance(ob1, (tuple, list)):
        if not isinstance(ob2, (tuple, list)):
            return '<type does not match>'
        if len(ob1) != len(ob2):
            return '<lengths do not match>'
        # Test all elements
        for i in range(len(ob1)):
            not_equal = _not_equal(ob1[i], ob2[i])
            if not_equal:
                return ('[%i]' % i) + not_equal
    
    elif isinstance(ob1, VirtualArray):
        if not isinstance(ob2, VirtualArray):
            return '<type does not match>'
        # Test properties
        if not (    ob1.shape==ob2.shape and
                    ob1.dtype==ob2.dtype and
                    ob1.data==ob2.data ):
            return '<array does not match>'
    
    elif np and isinstance(ob1, np.ndarray):
        if not isinstance(ob2, np.ndarray):
            return '<type does not match>'
        # Test properties
        if not (    ob1.shape==ob2.shape and
                    ob1.dtype==ob2.dtype and
                    (ob1==ob2).sum()==ob1.size ):
            return '<array does not match>'
    
    else:
        # Use default equality operator
        if not (ob1 == ob2):
            return '<objects not equal>'


def _isvalidname(name):
    """ _isvalidname(name)
    
    Returns attribute name, or None, if not valid
    
    """
    
    # Is it a string?
    if not ( name and isinstance(name, string_types) ):
        return None
    
    # Check name
    namechars = str('abcdefghijklmnopqrstuvwxyz_0123456789')
    name2 = name.lower()
    if name2[0] not in namechars[0:-10]:
        return None
    tmp = list(map(lambda x: x not in namechars, name2[2:]))
    
    # Return
    if sum(tmp)==0:
        return name


def _shapeString(ob):
    """ _shapeString(ob)
    
    Returns a string that represents the shape of the given array.
    
    """
    ss = str()
    for n in ob.shape:
        ss += '%ix' % n
    return ss[:-1]



class Struct(object):
    """ Struct(dictionary=None)
    
    Object to holds named data (syntactic sugar for a dictionary).
    
    Attributes can be any of the seven SSDF supported types:
    struct/dict, tuple/list, numpy array, (Unicode) string, int, float, None.
    
    Elements can be added in two ways:
        * s.foo = 'bar'       # the object way
        * s['foo'] = 'bar'    # the dictionary way
    
    Supported features
    ------------------
    * Iteration - yields the keys/names in the struct
    * len() - returns the number of elements in the struct
    * del statement can be used to remove elements
    * two structs can be added, yielding a new struct with combined elements
    * testing for equality with other structs
    
    Notes
    -----
      * The keys in the given dict should be valid names (invalid
        keys are ignoired).
      * On saving, names starting with two underscores are ignored.
      * This class does not inherit from dict to keep its namespace clean,
        avoid nameclashes, and to enable autocompletion of its items in
        most IDE's.
      * To get the underlying dict, simply use s.__dict__.
    
    """
    
    # Indentifier
    __is_ssdf_struct__ = True
    
        
    def __init__(self, a_dict=None):
        
        # Plain struct?
        if a_dict is None:
            return
        
        if not isinstance(a_dict, dict) and not isstruct(a_dict):
            tmp = "Struct can only be initialized with a Struct or a dict."
            raise ValueError(tmp)
        else:
            # Try loading from object
            
            def _getValue(val):
                """ Get the value, as suitable for Struct. """
                if isinstance(val, (string_types,) + _FLOAT_TYPES + _INT_TYPES ):
                    return val
                if np and isinstance(val, np.ndarray):
                    return val
                elif isinstance(val,(tuple,list)):
                    L = list()
                    for element in val:
                        L.append( _getValue(element) )
                    return L
                elif isinstance(val, dict):
                    return Struct(val)
                else:
                    pass # leave it
            
            # Copy all keys in the dict that are not methods
            for key in a_dict:
                if not _isvalidname(key):
                    print("Ignoring invalid key-name '%s'." % key)
                    continue
                val = a_dict[key]
                self[key] = _getValue(val)
    
    
    def __getitem__(self, key):
        # Name ok?
        key2 = _isvalidname(key)
        if not key2:
            raise KeyError("Trying to get invalid name '%s'." % key)
        # Name exists?
        if not key in self.__dict__:
            raise KeyError(str(key))
        # Return
        return self.__dict__[key]
    
    
    def __setitem__(self, key, value):
        # Name ok?
        key2 = _isvalidname(key)
        if not key2:
            raise KeyError("Trying to set invalid name '%s'." % key)
        # Set
        self.__dict__[key] = value
    
    
    def __iter__(self):
        """ Returns iterator over keys. """
        return self.__dict__.__iter__()
   
    
    def __delitem__(self, key):
        return self.__dict__.__delitem__(key)
    
    
    def __len__(self):
        """ Return amount of fields in the Struct object. """
        return len(self.__dict__)
    
    def __add__(self, other):
        """ Enable adding two structs by combining their elemens. """
        s = Struct()
        s.__dict__.update(self.__dict__)
        s.__dict__.update(other.__dict__)
        return s
    
    def __eq__(self, other):
        return not _not_equal(self, other)
    
    def __repr__(self):
        """ Short string representation. """
        return "<SSDF struct instance with %i elements>" % len(self)
    
    def __str__(self):
        """ Long string representation. """
        
        # Get alignment value
        c = 0
        for key in self:
            c = max(c, len(key))
        
        # How many chars left (to display on less than 80 lines)
        charsLeft = 79 - (c+4) # 2 spaces and ': '
        
        s = 'Elements in SSDF struct:\n'
        for key in self:
            if key.startswith("__"):
                continue
            tmp = "%s" % (key)
            value = self[key]
            valuestr = repr(value)
            if len(valuestr)>charsLeft or '\n' in valuestr:
                typestr = str(type(value))[7:-2]
                if np and isinstance(value,np.ndarray):
                    shapestr = _shapeString(value)
                    valuestr = "<array %s %s>" %(shapestr,str(value.dtype))
                elif isinstance(value, string_types):
                    valuestr = valuestr[:charsLeft-3] + '...'
                    #valuestr = "<string with length %i>" % (typestr, len(value))
                else:
                    valuestr = "<%s with length %i>" % (typestr, len(value))
            s += tmp.rjust(c+2) + ": %s\n" % (valuestr)
        return s


class VirtualArray(object):
    """ VirtualArray
    
    A VirtualArray represents an array when numpy is not available.
    This enables preserving the array when saving back a loaded dataset.
    
    """
    def __init__(self, shape, dtype, data):
        self.shape = tuple(shape)
        self.dtype = dtype
        self.data = data
    
    def tostring(self):
        return self.data
    
    @property
    def size(self):
        if self.shape:
            return reduce( lambda a,b:a*b, self.shape)
        else:
            return 1


class SSDFReader:
    
    def build_tree(self, root, blocks):
        """ build_tree(root, blocks)
        
        Build up the tree using the indentation information in the blocks.
        The tree is build up from the given root.
        
        """
        tree = [root]
        for block in blocks:
            # Select leaf in tree
            while block._indent <= tree[-1]._indent:
                tree.pop()
            # Append (to object and to simple tree structure)
            tree[-1]._children.append(block)
            tree.append(block)

    def serialize_struct(self, object, f=None):
        raise NotImplementedError()
    
    def read(self, file_or_string):
        raise NotImplementedError()


class SSDFWriter:
 
    def flatten_tree(self, block, sort=False):
        """ flatten_tree(block, sort=False)
        
        Returns a flat list containing the given block and
        all its children.
        
        If sort is True, packs blocks such that the data
        structures consisting of less blocks appear first.
        
        """
        
        # Get list of strings for each child
        listOfLists = []
        for child in block._children:
            childList = self.flatten_tree(child, sort)
            listOfLists.append( childList )
        
        # Sort by length
        if sort and listOfLists and block._type == _TYPE_DICT:
            listOfLists.sort(key=len)
        
        # Produce flat list
        flatList = [block]
        for childList in listOfLists:
            flatList.extend(childList)
        
        # Done
        return flatList

    def write(self, object, f=None):
        raise NotImplementedError()



class Block:
    """ Block
    
    A block represents a data element. This is where the conversion from
    Python objects to text/bytes and vice versa occurs.
    
    A block is a line in a text file or a piece of data in a binary file.
    A block contains all information about indentation, name, and value
    of the data element that it represents. The raw representation of its
    value is refered to as 'data'.
    
    """
    
    def __init__(self, indent, blocknr, name=None, type=None, data=None):
        self._indent = indent
        self._blocknr = blocknr # for producing usefull read error messages
        
        self._name = name
        self._type = type # used by binary only (and text-dict)
        self._data = data # the raw data, bytes or string
        
        self._children = [] # used only by dicts and lists
    
    
    @classmethod
    def from_object(cls, indent, name, value):
        
        # Instantiate a block
        self = cls(indent, -1, name)
        
        # Set object's data
        if value is None:
            self._from_none()
        elif ClassManager.is_registered_class(value.__class__):
            s = value.__to_ssdf__()
            s[_CLASS_NAME] = value.__class__.__name__
            self._from_dict(s)
        elif isinstance(value, _INT_TYPES):
            self._from_int(value)
        elif isinstance(value, _FLOAT_TYPES):
            self._from_float(value)
        elif isinstance(value, bool):
            self._from_int(int(value))
        elif isinstance(value, string_types):
            self._from_unicode(value)
        elif np and isinstance(value, np.ndarray):
            self._from_array(value)
        elif isinstance(value, VirtualArray):
            self._from_array(value)
        elif isinstance(value, dict) or isstruct(value):
            self._from_dict(value)
        elif isinstance(value, (list, tuple)):
            self._from_list(value)
        else:
            # We do not know
            self._from_none()
            tmp = repr(value)
            if len(tmp) > 64:
                tmp = tmp[:64] + '...'
            if name is not None:
                print("SSDF: %s is unknown object: %s %s" %
                                    (name, tmp, repr(type(value)) ))
            else:
                print("SSDF: unknown object: %s %s" %
                                    (tmp, repr(type(value)) ))
        
        # Done
        return self



#!/usr/bin/env python
""" This example demonstrates rendering a color volume.
This example demonstrates two render styles. Note that
all render styles are capable of rendering color data.
"""

import numpy as np
import visvis as vv
app = vv.use()

# Use vv.aVolume to create random bars for each color plane
N = 64
vol = np.empty((N,N,N,3), dtype='float32')
for i in range(3):
    vol[:,:,:,i] = vv.aVolume(10,N)

# Show
vv.figure()
a1 = vv.subplot(121)
t1 = vv.volshow(vol[:,:,:,:], renderStyle = 'mip')
vv.title('color MIP render')
a2 = vv.subplot(122)
t2 = vv.volshow(vol[:,:,:,:], renderStyle = 'iso')
t2.isoThreshold = 0.5
vv.title('color ISO-surface render')

# Share cameras
a1.camera = a2.camera

# Run app
app.Run()



# -*- coding: utf-8 -*-
# flake8: noqa
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" Module wobjects

All wobjects are inserted in this namespace, thereby providing
the user with a list of all wobjects. All wobjects are also
inserted in the root visvis namespace.

"""

from visvis.core import (   Wobject, Line, PolarLine, Text )

from visvis.wobjects.textures import BaseTexture, Texture2D, Texture3D
from visvis.wobjects.textures import MotionTexture2D, MotionTexture3D
from visvis.wobjects.sliceTextures import SliceTexture, SliceTextureProxy
from visvis.wobjects.polygonalModeling import Mesh, OrientableMesh
from visvis.wobjects.motion import MotionDataContainer, MotionMixin, MotionSyncer



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" Module base

Defines the Wibject and Wobject classes, and the Position class.

"""

import OpenGL.GL as gl
import numpy as np

import weakref

from visvis.core import misc
from visvis.core.misc import basestring
from visvis.core.misc import (Transform_Base, Transform_Translate,
                                    Transform_Scale, Transform_Rotate)
from visvis.core import events
from visvis.utils.pypoints import Pointset, Quaternion, is_Point


# Define draw modes
DRAW_NORMAL = 1     # draw normally.
DRAW_FAST = 2       # draw like normal, but faster (while interacting)
DRAW_SHAPE = 3      # draw the spape of the object in the given color
DRAW_SCREEN = 4     # for wobjects to draw in screen coordinates


class BaseObject(object):
    """ BaseObject(parent)
    
    The base class for wibjects and wobjects.
    Instances of classes inherited from this class represent
    something that can be drawn.
    
    Wibjects and wobjects can have children and have a parent
    (which can be None in which case they are in orphan and never
    drawn). To change the structure, use the ".parent" property.
    They also can be set visible/invisible using the property ".visible".
    
    """
    
    def __init__(self, parent):
        # wheter or not the object has been destroyed
        self._destroyed = False
        # whether or not to draw the object
        self._visible = True
        # whether the object is currently being drawn
        self._isbeingdrawn = False
        # whether the object should draw its shape
        self._hitTest = False
        # the parent object
        self._parent = None
        # the children of this object
        self._children = []
        # the id of this object (can change on every draw)
        # used to determine over which object the mouse is.
        self._id = 0
        # a variable to indicate whether the mouse was pressed down here
        self._mousePressedDown = False
        
        # set parent
        self.parent = parent
        
        # create events
        self._eventMouseDown = events.EventMouseDown(self)
        self._eventMouseUp = events.EventMouseUp(self)
        self._eventDoubleClick = events.EventDoubleClick(self)
        self._eventEnter = events.EventEnter(self)
        self._eventLeave = events.EventLeave(self)
        #
        self._eventMotion = events.EventMotion(self)
        self._eventScroll = events.EventScroll(self)
        self._eventKeyDown = events.EventKeyDown(self)
        self._eventKeyUp = events.EventKeyUp(self)
    
    
    @property
    def eventMouseDown(self):
        """ Fired when the mouse is pressed down on this object. (Also
        fired the first click of a double click.)
        """
        return self._eventMouseDown
    @property
    def eventMouseUp(self):
        """ Fired when the mouse is released after having been clicked down
        on this object (even if the mouse is now not over the object). (Also
        fired on the first click of a double click.)
        """
        return self._eventMouseUp
    @property
    def eventDoubleClick(self):
        """ Fired when the mouse is double-clicked on this object.
        """
        return self._eventDoubleClick
    @property
    def eventEnter(self):
        """ Fired when the mouse enters this object or one of its children.
        """
        return self._eventEnter
    @property
    def eventLeave(self):
        """ Fired when the mouse leaves this object (and is also not over any
        of it's children).
        """
        return self._eventLeave
    
    @property
    def eventMotion(self):
        """ Fires when the mouse is moved over the object. Not fired when
        the mouse is over one of its children.
        """
        return self._eventMotion
    
    @property
    def eventScroll(self):
        """ Fires when the scroll wheel is used while over the object.
        Not fired when the mouse is over one of its children.
        """
        return self._eventScroll
    
    @property
    def eventKeyDown(self):
        """ Fires when the mouse is moved over the object. Not fired when
        the mouse is over one of its children.
        """
        return self._eventKeyDown
    @property
    def eventKeyUp(self):
        """ Fires when the mouse is moved over the object. Not fired when
        the mouse is over one of its children.
        """
        return self._eventKeyUp
    
    
    def _testWhetherShouldDrawShape(self):
        """ Tests whether any of the events has handlers registered
        to it. If so, this object should draw its shape.
        This method is called by the event objects when handlers are
        added or removed.
        """
        self._hitTest = False
        for name in dir(self.__class__):
            if name.startswith('event'):
                event = getattr(self, name, None)
                if event and event.hasHandlers:
                    self._hitTest = True
                    break
    
    
    def _DrawTree(self, mode=DRAW_NORMAL, pickerHelper=None):
        """ Draw the wibject/wobject and all of its children.
        """
        
        # are we alive
        if self._destroyed:
            print("Warning, cannot draw destroyed object: %s" % str(self))
            return
        
        # only draw if visible
        if not self.visible:
            return
        
        # transform
        gl.glPushMatrix()
        self._Transform()
        
        # draw self
        self._isbeingdrawn = True
        try:
            if mode==DRAW_SHAPE:
                if self._hitTest:
                    clr = pickerHelper.GetColorFromId(self._id)
                    self.OnDrawShape(clr)
            elif mode==DRAW_SCREEN:
                self.OnDrawScreen()
            elif mode==DRAW_FAST:
                self.OnDrawFast()
            elif mode==DRAW_NORMAL:
                self.OnDraw()
            else:
                raise Exception("Invalid mode for _DrawTree.")
            
            # draw children
            for item in self._children:
                if hasattr(item, '_DrawTree'):
                    item._DrawTree(mode,pickerHelper)
        
        finally:
            self._isbeingdrawn = False
        
        # transform back
        gl.glPopMatrix()
    
    
    def Destroy(self, setContext=True):
        """ Destroy()
        
        Destroy the object.
          * Removes itself from the parent's children
          * Calls Destroy() on all its children
          * Calls OnDestroyGl and OnDestroy on itself
        
        Note1: do not overload, overload OnDestroy().
        Note2: it's best not to reuse destroyed objects. To temporary disable
        an object, better use "ob.parent=None", or "ob.visible=False".
        
        """
        
        # Post draw event
        self.Draw()
        
        # We must make this the current OpenGL context because OnDestroy
        # methods of objects may want to remove textures etc.
        # When you do not do this, you can get really weird bugs.
        # This works nice, the children will not need to do this,
        # as when they try, THIS object is already detached and fig is None.
        if setContext:
            fig = self.GetFigure()
            if fig:
                fig._SetCurrent()
        
        # Destroy children. This will unwind to the leafs of the tree, and
        # thus call OnDestroy() on childless objects only. This means the
        # parent of all objects remain intact, which can be necessary for
        # some objects to clean up nicely.
        for child in self.children:
            child.Destroy(False)
        
        # Actual destroy
        self.OnDestroyGl()
        self.OnDestroy()
        self._destroyed = True
        
        # Leave home (using the property causes recursion)
        if hasattr(self._parent, '_children'):
            while self in self._parent._children:
                self._parent._children.remove(self)
        if hasattr(self._parent, '_wobjects'):
            while self in self._parent._wobjects:
                self._parent._wobjects.remove(self)
        self._parent = None
    
    
    def DestroyGl(self, setContext=True):
        """ DestroyGl()
        
        Destroy the OpenGl objects managed by this object.
          * Calls DestroyGl() on all its children.
          * Calls OnDestroyGl() on itself.
        
        Note: do not overload, overload OnDestroyGl().
        
        """
        
        # make the right openGlcontext current
        if setContext:
            fig = self.GetFigure()
            if fig:
                fig._SetCurrent()
        
        # let children clean up their openGl stuff
        for child in self._children:
            child.DestroyGl(False)
        
        # Clean up our own bits
        self.OnDestroyGl()
    
    
    def __del__(self):
        self.Destroy()
    
    
    def _Transform(self):
        """ Add transformations to modelview matrix such that the object
        is displayed properly.
        """
        pass # implemented diffently by the wibject and wobject class
    
    
    def OnDraw(self):
        """ OnDraw()
        Perform the opengl commands to draw this wibject/wobject.
        Objects should overload this method to draw themselves.
        """
        pass
    
    def OnDrawFast(self):
        """ OnDrawFast()
        
        Overload this to provide a faster version to draw (but
        less pretty), which is called when the scene is zoomed/translated.
        By default, this calls OnDraw()
        
        """
        self.OnDraw()
    
    def OnDrawShape(self, color):
        """ OnDrawShape(color)
        
        Perform  the opengl commands to draw the shape of the object
        in the given color.
        If not implemented, the object cannot be picked.
        
        """
        pass
    
    def OnDrawScreen(self):
        """ OnDrawScreen()
        
        Draw in screen coordinates. To be used for wobjects that
        need drawing in screen coordinates (like text). Wibjects are
        always drawn in screen coordinates (using OnDraw).
        
        """
        pass
        
    def OnDestroy(self):
        """ OnDestroy()
        
        Overload this to clean up any resources other than the GL objects.
        
        """
        for att in list(self.__dict__.values()):
            if isinstance(att, events.BaseEvent):
                att.Unbind()
             
    
    def OnDestroyGl(self):
        """ OnDestroyGl()
        
        Overload this to clean up any OpenGl resources.
        
        """
        pass
    
    @misc.PropWithDraw
    def visible():
        """ Get/Set whether the object should be drawn or not.
        If set to False, the hittest is also not performed.
        """
        def fget(self):
            return self._visible
        def fset(self, value):
            self._visible = bool(value)
        return locals()
    
    
    @misc.Property
    def hitTest():
        """ Get/Set whether mouse events are generated for this object.
        From v1.7 this property is set automatically, and need not be set
        to receive mouse events.
        """
        def fget(self):
            return self._hitTest
        def fset(self, value):
            self._hitTest = bool(value)
        return locals()
    
    
    @misc.PropWithDraw
    def parent():
        """ Get/Set the parent of this object. Use this to change the
        tree structure of your visualization objects (for example move a line
        from one axes to another).
        """
        def fget(self):
            return self._parent
        def fset(self,value):
            
            # Post draw event at parent
            self.Draw()
            
            # init lists to update
            parentChildren = None
            if hasattr(value, '_children'):
                parentChildren = value._children
            
            # check if this is a valid parent
            tmp = "Cannot change to that parent, "
            if value is None:
                # an object can be an orphan
                pass
            elif value is self:
                # an object cannot be its own parent
                raise TypeError(tmp+"because that is the object itself!")
            elif isinstance(value, Wibject):
                # some wibject parents can hold wobjects
                if isinstance(self, Wobject):
                    if hasattr(value, '_wobjects'):
                        parentChildren = value._wobjects
                    else:
                        tmp2 = "a wobject can only have a wibject-parent "
                        raise TypeError(tmp+tmp2+"if it can hold wobjects!")
            elif isinstance(value, Wobject):
                # a wobject can only hold wobjects
                if isinstance(self, Wibject):
                    raise TypeError(tmp+"a wibject cant have a wobject-parent!")
            else:
                raise TypeError(tmp+"it is not a wibject or wobject or None!")
            
            # remove from parents childrens list
            if hasattr(self._parent, '_children'):
                while self in self._parent._children:
                    self._parent._children.remove(self)
            if hasattr(self._parent, '_wobjects'):
                while self in self._parent._wobjects:
                    self._parent._wobjects.remove(self)
            
            # Should we destroy GL objects (because we are removed
            # from an OpenGL context)?
            figure1 = self.GetFigure()
            figure2 = None
            if hasattr(value, 'GetFigure'):
                figure2 = value.GetFigure()
            if figure1 and (figure1 is not figure2):
                self.DestroyGl()
            
            # set and add to new parent
            self._parent = value
            if parentChildren is not None:
                parentChildren.append(self)
        
        return locals()
    
    
    @property
    def children(self):
        """ Get a shallow copy of the list of children.
        """
        return [child for child in self._children]
    
    
    def GetFigure(self):
        """ GetFigure()
        
        Get the figure that this object is part of.
        The figure represents the OpenGL context.
        Returns None if it has no figure.
        
        """
        # init
        iter = 0
        object = self
        # search
        while hasattr(object,'parent'):
            iter +=1
            if object.parent is None:
                break
            if iter > 100:
                break
            object = object.parent
        # check
        if object.parent is None and hasattr(object, '_SwapBuffers'):
            return object
        else:
            return None
    
    
    def Draw(self, fast=False):
        """ Draw(fast=False)
        
        For wibjects: calls Draw() on the figure that contains this object.
        For wobjects: calls Draw() on the axes that contains this object.
        
        """
        if self._isbeingdrawn:
            return False
        else:
            fig = self.GetFigure()
            if fig:
                fig.Draw()
            return True
    
    
    def FindObjects(self, spec):
        """ FindObjects(pattern)
        
        Find the objects in this objects' children, and its childrens
        children, etc, that correspond to the given pattern.
        
        The pattern can be a class or tuple of classes, an attribute name
        (as a string) that the objects should have, or a callable that
        returns True or False given an object. For example
        'lambda x: ininstance(x, cls)' will do the same as giving a class.
        
        If 'self' is a wibject and has a _wobject property (like the Axes
        wibject) this method also performs the search in the list of wobjects.
        
        """
        
        
        # Parse input
        if hasattr(spec, 'func_name'):
            callback = spec
        elif isinstance(spec, (type, tuple)):
            callback = lambda x: isinstance(x, spec)
        elif isinstance(spec, basestring):
            callback = lambda x: hasattr(x, spec)
        elif hasattr(spec, '__call__'):
            callback = spec # other callable
        else:
            raise ValueError('Invalid argument for FindObjects')
        
        # Init list with result
        result = []
        
        # Try all children recursively
        for child in self._children:
            if callback(child):
                result.append(child)
            result.extend( child.FindObjects(callback) )
        if hasattr(self, '_wobjects'):
            for child in self._wobjects:
                if callback(child):
                    result.append(child)
                result.extend( child.FindObjects(callback) )
        
        # Done
        return result
    
    
    def GetWeakref(self):
        """ GetWeakref()
        
        Get a weak reference to this object.
        Call the weakref to obtain the real reference (or None if it's dead).
        
        """
        return weakref.ref( self )


class Wibject(BaseObject):
    """ Wibject(parent)
    
    A Wibject (widget object) is a 2D object drawn in
    screen coordinates. A Figure is a widget and so are an Axes and a
    PushButton. Wibjects have a position property to set their location
    and size. They also have a background color and multiple event properties.
    
    This class may also be used as a container object for other wibjects.
    An instance of this class has no visual appearance. The Box class
    implements drawing a rectangle with an edge.
    
    """
    
    def __init__(self, parent):
        BaseObject.__init__(self, parent)
        
        # the position of the widget within its parent
        self._position = Position( 10,10,50,50, self)
        
        # colors and edge
        self._bgcolor = (0.8,0.8,0.8)
        
        # event for position
        self._eventPosition = events.EventPosition(self)
    
    
    @property
    def eventPosition(self):
        """ Fired when the position (or size) of this wibject changes.
        """
        return self._eventPosition
    
    
    @misc.PropWithDraw
    def position():
        """ Get/Set the position of this wibject. Setting can be done
        by supplying either a 2-element tuple or list to only change
        the location, or a 4-element tuple or list to change location
        and size.
        
        See the docs of the vv.base.Position class for more information.
        """
        def fget(self):
            return self._position
        def fset(self, value):
            self._position.Set(value)
        return locals()
    
    
    @misc.PropWithDraw
    def bgcolor():
        """ Get/Set the background color of the wibject.
        """
        def fget(self):
            return self._bgcolor
        def fset(self, value):
            self._bgcolor = misc.getColor(value, 'setting bgcolor')
        return locals()
    
    
    def _Transform(self):
        """ _Transform()
        Apply a translation such that the wibject is
        drawn in the correct place.
        """
        # skip if we are on top
        if not self.parent:
            return
        # get posision in screen coordinates
        pos = self.position
        # apply
        gl.glTranslatef(pos.left, pos.top,0.0)

    
    def OnDrawShape(self, clr):
        # Implementation of the OnDrawShape method.
        gl.glColor(clr[0], clr[1], clr[2], 1.0)
        w,h = self.position.size
        gl.glBegin(gl.GL_POLYGON)
        gl.glVertex2f(0,0)
        gl.glVertex2f(0,h)
        gl.glVertex2f(w,h)
        gl.glVertex2f(w,0)
        gl.glEnd()



class Wobject(BaseObject):
    """ Wobject(parent)
    
    A Wobject (world object) is a visual element that
    is drawn in 3D world coordinates (in the scene). Wobjects can be
    children of other wobjects or of an Axes object (which is the
    wibject that represents the scene).
    
    To each wobject, several transformations can be applied,
    which are also applied to its children. This way complex models can
    be build. For example, in a robot arm the fingers would be children
    of the hand, so that when the hand moves or rotates, the fingers move
    along automatically. The fingers can then also be moved without affecting
    the hand or other fingers.
    
    The transformations are represented by Transform_* objects in
    the list named "transformations". The transformations are applied
    in the order as they appear in the list.
    
    """
    
    def __init__(self, parent):
        BaseObject.__init__(self, parent)
        
        # the transformations applied to the object
        self._transformations = []
    
    
    @property
    def transformations(self):
        """ Get the list of transformations of this wobject. These
        can be Transform_Translate, Transform_Scale, or Transform_Rotate
        instances.
        """
        return self._transformations
    
    
    def GetAxes(self):
        """ GetAxes()
        
        Get the axes in which this wobject resides.
        
        Note that this is not necesarily an Axes instance (like the line
        objects in the Legend wibject).
        
        """
        par = self.parent
        if par is None:
            return None
        while not isinstance(par, Wibject):
            par = par.parent
            if par is None:
                return None
        return par
    
    
    def Draw(self, fast=False):
        """ Draw(fast=False)
        
        Calls Draw on the axes that contains this object.
        
        """
        if self._isbeingdrawn:
            return False
        else:
            axes = self.GetAxes()
            if axes:
                axes.Draw()
            return True
    
    
    def _GetLimits(self, *args):
        """ _GetLimits(self, x1, x2, y1, y2, z1, z2)
        
        Get the limits in world coordinates between which the object
        exists. This is used by the Axes class to set the camera correctly.
        If None is returned, the limits are undefined.
        
        Inheriting Wobject classes should overload this method. However, they
        can use this method to take all transformations into account by giving
        the cornerpoints of the untransformed object.
        
        Returns a 3 element tuple of vv.Range instances: xlim, ylim, zlim.
        
        """
        
        # Examine args
        if not args:
            minx, maxx, miny, maxy, minz, maxz = [], [], [], [], [], []
        elif len(args) == 6:
            minx, maxx, miny, maxy, minz, maxz = tuple([[arg] for arg in args])
        else:
            raise ValueError("_Getlimits expects 0 or 6 arguments.")
        
        # Get limits of children
        for ob in self.children:
            tmp = ob._GetLimits()
            if tmp is not None:
                limx, limy, limz = tmp
                minx.append(limx.min); maxx.append(limx.max)
                miny.append(limy.min); maxy.append(limy.max)
                minz.append(limz.min); maxz.append(limz.max)
        
        # Do we have limits?
        if not (minx and maxx and miny and maxy and minz and maxz):
            return None
        
        # Take min and max
        x1, y1, z1 = tuple([min(val) for val in [minx, miny, minz]])
        x2, y2, z2 = tuple([max(val) for val in [maxx, maxy, maxz]])
        
        # Make pointset of eight cornerpoints
        pp = Pointset(3)
        for x in [x1, x2]:
            for y in [y1, y2]:
                for z in [z1, z2]:
                    pp.append(x,y,z)
        
        # Transform these points
        for i in range(len(pp)):
            pp[i] = self.TransformPoint(pp[i], self)
        
        # Return limits
        xlim = misc.Range( pp[:,0].min(), pp[:,0].max() )
        ylim = misc.Range( pp[:,1].min(), pp[:,1].max() )
        zlim = misc.Range( pp[:,2].min(), pp[:,2].max() )
        return xlim, ylim, zlim
    
    
    def TransformPoint(self, p, baseWobject=None):
        """ TransformPoint(p, baseWobject=None)
        
        Transform a point in the local coordinate system of this wobject
        to the coordinate system of the given baseWobject (which should be
        a parent of this wobject), or to the global (Axes) coordinate
        system if not given.
        
        This is done by taking into account the transformations applied
        to this wobject and its parent wobjects.
        
        If baseWobject is the current wobject itself, only the tranformations
        of this wobject are applied.
        
        """
        if not (is_Point(p) and p.ndim==3):
            raise ValueError('TransformPoint only accepts a 3D point')
        
        # Init wobject as itself. Next round it will be its parent, etc.
        wobject = self
        
        # Iterate over wobjects until we reach the Axes or None
        while isinstance(wobject, Wobject):
            # Iterate over all transformations
            for t in reversed(wobject._transformations):
                if isinstance(t, Transform_Translate):
                    p.x += t.dx
                    p.y += t.dy
                    p.z += t.dz
                elif isinstance(t, Transform_Scale):
                    p.x *= t.sx
                    p.y *= t.sy
                    p.z *= t.sz
                elif isinstance(t, Transform_Rotate):
                    angle = float(t.angle * np.pi / 180.0)
                    q = Quaternion.create_from_axis_angle(angle, t.ax, t.ay, t.az)
                    p = q.rotate_point(p)
            # Done or move to next parent?
            if wobject is baseWobject:
                break
            else:
                wobject = wobject.parent
        
        # Done
        return p
    
    
    def _Transform(self):
        """ _Transform()
        Apply all listed transformations of this wobject.
        """
        for t in self.transformations:
            if not isinstance(t, Transform_Base):
                continue
            elif isinstance(t, Transform_Translate):
                gl.glTranslate(t.dx, t.dy, t.dz)
            elif isinstance(t, Transform_Scale):
                gl.glScale(t.sx, t.sy, t.sz)
            elif isinstance(t, Transform_Rotate):
                gl.glRotate(t.angle, t.ax, t.ay, t.az)


class Position(object):
    """ Position(x,y,w,h, wibject_instance)
    
    The position class stores and manages the position of wibjects. Each
    wibject has one Position instance associated with it, which can be
    obtained (and updated) using its position property.
    
    The position is represented using four values: x, y, w, h. The Position
    object can also be indexed to get or set these four values.
    
    Each element (x,y,w,h) can be either:
      * The integer amount of pixels relative to the wibjects parent's position.
      * The fractional amount (float value between 0.0 and 1.0) of the parent's width or height.
    
    Each value can be negative. For x and y this simply means a negative
    offset from the parent's left and top. For the width and height the
    difference from the parent's full width/height is taken.
    
    An example: a position (-10, 0.5, 150,-100), with a parent's size of
    (500,500) is equal to (-10, 250, 150, 400) in pixels.
    
    Remarks:
      * fractional, integer and negative values may be mixed.
      * x and y are considered fractional on <-1, 1>
      * w and h are considered fractional on [-1, 1]
      * the value 0 can always be considered to be in pixels
    
    The position class also implements several "long-named" properties that
    express the position in pixel coordinates. Internally a version in pixel
    coordinates is buffered, which is kept up to date. These long-named
    (read-only) properties are:
    left, top, right, bottom, width, height,
    
    Further, there are a set of properties which express the position in
    absolute coordinates (not relative to the wibject's parent):
    absLeft, absTop, absRight, absBottom
    
    Finally, there are properties that return a two-element tuple:
    topLeft, bottomRight, absTopLeft, absBottomRight, size
    
    The method InPixels() returns a (copy) Position object which represents
    the position in pixels.
    
    """
    
    def __init__(self, x, y, w, h, owner):
        
        # test owner
        if not isinstance(owner , Wibject):
            raise ValueError('A positions owner can only be a wibject.')
        
        # set
        self._x, self._y, self._w, self._h = x, y, w, h
        
        # store owner using a weak reference
        self._owner = weakref.ref(owner)
        
        # init position in pixels and absolute (as a tuples)
        self._inpixels = None
        self._absolute = None
        
        # do not _update() here, beacause the owner will not have assigned
        # this object to its _position attribute yet.
        
        # but we can calculate our own pixels
        self._CalculateInPixels()
    
    
    def Copy(self):
        """ Copy()
        
        Make a copy of this position instance.
        
        """
        p = Position(self._x, self._y, self._w, self._h, self._owner())
        p._inpixels = self._inpixels
        p._absolute = self._absolute
        return p
    
    
    def InPixels(self):
        """ InPixels()
        
        Return a copy, but in pixel coordinates.
        
        """
        p = Position(self.left,self.top,self.width,self.height, self._owner())
        p._inpixels = self._inpixels
        p._absolute = self._absolute
        return p
    
    
    def __repr__(self):
        return "<Position %1.2f, %1.2f,  %1.2f, %1.2f>" % (
            self.x, self.y, self.w, self.h)
    
    
    ## For keeping _inpixels up-to-date
    
    
    def _Update(self):
        """ _Update()
        
        Re-obtain the position in pixels. If the obtained position
        differs from the current position-in-pixels, _Changed()
        is called.
        
        """
        
        # get old version, obtain and store new version
        ip1 = self._inpixels + self._absolute
        self._CalculateInPixels()
        ip2 = self._inpixels + self._absolute
        
        if ip2:
            if ip1 != ip2: # also if ip1 is None
                self._Changed()
    
    
    def _Changed(self):
        """ _Changed()
        
        To be called when the position was changed.
        Will fire the owners eventPosition and will call
        _Update() on the position objects of all the owners
        children.
        
        """
        # only notify if this is THE position of the owner (not a copy)
        owner = self._owner()
        if owner and owner._position is self:
            if hasattr(owner, 'eventPosition'):
                owner.eventPosition.Fire()
                #print('firing position event for', owner)
            for child in owner._children:
                if hasattr(child, '_position'):
                    child._position._Update()
    
    
    def _GetFractionals(self):
        """ Get a list which items are considered relative.
        Also int()'s the items which are not.
        """
        # init
        fractionals = [0,0,0,0]
        # test
        for i in range(2):
            if self[i] > -1 and self[i] < 1 and self[i]!=0:
                fractionals[i] = 1
        for i in range(2,4):
            if self[i] >= -1 and self[i] <= 1 and self[i]!=0:
                fractionals[i] = 1
        # return
        return fractionals
    
    
    def _CalculateInPixels(self):
        """ Return the position in screen coordinates as a tuple.
        """
        
        # to test if this is easy
        fractionals = self._GetFractionals()
        negatives = [int(self[i]<0) for i in range(4)]
        
        # get owner
        owner = self._owner()
        
        # if owner is a figure, it cannot have relative values
        if hasattr(owner, '_SwapBuffers'):
            self._inpixels = (self._x, self._y, self._w, self._h)
            self._absolute = self._inpixels
            return
        
        # test if we can calculate
        if not isinstance(owner, Wibject):
            raise Exception("Can only calculate the position in pixels"+
                            " if the position instance is owned by a wibject!")
        # else, the owner must have a parent...
        if owner.parent is None:
            print(owner)
            raise Exception("Can only calculate the position in pixels"+
                            " if the owner has a parent!")
        
        # get width/height of parent
        ppos = owner.parent.position
        whwh = ppos.width, ppos.height
        whwh = (whwh[0], whwh[1], whwh[0], whwh[1])
        
        # calculate!
        pos = [self._x, self._y, self._w, self._h]
        if max(fractionals)==0 and max(negatives)==0:
            pass # no need to calculate
        else:
            for i in range(4):
                if fractionals[i]:
                    pos[i] = pos[i]*whwh[i]
                if i>1 and negatives[i]:
                    pos[i] = whwh[i] + pos[i]
                # make sure it's int (even if user supplied floats > 1)
                pos[i] = int(pos[i])
        
        # abs pos is based on the inpixels version, but x,y corrected.
        apos = [p for p in pos]
        if ppos._owner().parent:
            apos[0] += ppos.absLeft
            apos[1] += ppos.absTop
        
        # store
        self._inpixels = tuple(pos)
        self._absolute = tuple(apos)
    
    
    ## For getting and setting
    
    @misc.DrawAfter
    def Set(self, *args):
        """ Set(*args)
        
        Set(x, y, w, h) or Set(x, y).
        
        """
        
        # if tuple or list was given
        if len(args)==1 and hasattr(args[0],'__len__'):
            args = args[0]
        
        # apply
        if len(args)==2:
            self._x = args[0]
            self._y = args[1]
        elif len(args)==4:
            self._x = args[0]
            self._y = args[1]
            self._w = args[2]
            self._h = args[3]
        else:
            raise ValueError("Invalid number of arguments to position.Set().")
        
        # we need an update now
        self._Update()
    
    
    @misc.DrawAfter
    def Correct(self, dx=0, dy=0, dw=0, dh=0):
        """ Correct(dx=0, dy=0, dw=0, dh=0)
        
        Correct the position by suplying a delta amount of pixels.
        The correction is only applied if the attribute is in pixels.
        
        """
        
        # get fractionals
        fractionals = self._GetFractionals()
        
        # apply correction if we can
        if dx and not fractionals[0]:
            self._x += int(dx)
        if dy and not fractionals[1]:
            self._y += int(dy)
        if dw and not fractionals[2]:
            self._w += int(dw)
        if dh and not fractionals[3]:
            self._h += int(dh)
        
        # we need an update now
        self._Update()
    
    
    def __getitem__(self,index):
        if not isinstance(index,int):
            raise IndexError("Position only accepts single indices!")
        if index==0: return self._x
        elif index==1: return self._y
        elif index==2: return self._w
        elif index==3: return self._h
        else:
            raise IndexError("Position only accepts indices 0,1,2,3!")
    
    
    def __setitem__(self,index, value):
        if not isinstance(index,int):
            raise IndexError("Position only accepts single indices!")
        if index==0: self._x = value
        elif index==1: self._y = value
        elif index==2: self._w = value
        elif index==3: self._h = value
        else:
            raise IndexError("Position only accepts indices 0,1,2,3!")
        # we need an update now
        self._Update()
    
    
    def Draw(self):
        # Redraw owner
        owner = self._owner()
        if owner is not None:
            owner.Draw()
    
    
    @misc.PropWithDraw
    def x():
        """ Get/Set the x-element of the position. This value can be
        an integer value or a float expressing the x-position as a fraction
        of the parent's width. The value can also be negative.
        """
        def fget(self):
            return self._x
        def fset(self,value):
            self._x = value
            self._Update()
        return locals()
    
    @misc.PropWithDraw
    def y():
        """ Get/Set the y-element of the position. This value can be
        an integer value or a float expressing the y-position as a fraction
        of the parent's height. The value can also be negative.
        """
        def fget(self):
            return self._y
        def fset(self,value):
            self._y = value
            self._Update()
        return locals()
    
    @misc.PropWithDraw
    def w():
        """ Get/Set the w-element of the position. This value can be
        an integer value or a float expressing the width as a fraction
        of the parent's width. The value can also be negative, in which
        case it's subtracted from the parent's width.
        """
        def fget(self):
            return self._w
        def fset(self,value):
            self._w = value
            self._Update()
        return locals()
    
    @misc.PropWithDraw
    def h():
        """ Get/Set the h-element of the position. This value can be
        an integer value or a float expressing the height as a fraction
        of the parent's height. The value can also be negative, in which
        case it's subtracted from the parent's height.
        """
        def fget(self):
            return self._h
        def fset(self,value):
            self._h = value
            self._Update()
        return locals()
    
    ## Long names properties expressed in pixels
    
    @property
    def left(self):
        """ Get the x-element of the position, expressed in pixels.
        """
        tmp = self._inpixels
        return tmp[0]
    
    @property
    def top(self):
        """ Get the y-element of the position, expressed in pixels.
        """
        tmp = self._inpixels
        return tmp[1]
    
    @property
    def width(self):
        """ Get the w-element of the position, expressed in pixels.
        """
        tmp = self._inpixels
        return tmp[2]
    
    @property
    def height(self):
        """ Get the h-element of the position, expressed in pixels.
        """
        tmp = self._inpixels
        return tmp[3]
    
    @property
    def right(self):
        """ Get left+width.
        """
        tmp = self._inpixels
        return tmp[0] + tmp[2]
    
    @property
    def bottom(self):
        """ Get top+height.
         """
        tmp = self._inpixels
        return tmp[1] + tmp[3]
    
    @property
    def topLeft(self):
        """ Get a tuple (left, top).
        """
        tmp = self._inpixels
        return tmp[0], tmp[1]
    
    @property
    def bottomRight(self):
        """ Get a tuple (right, bottom).
        """
        tmp = self._inpixels
        return tmp[0] + tmp[2], tmp[1] + tmp[3]
    
    @property
    def size(self):
        """ Get a tuple (width, height).
        """
        tmp = self._inpixels
        return tmp[2], tmp[3]
    
    ## More long names for absolute position
    
    @property
    def absLeft(self):
        """ Get the x-element of the position, expressed in absolute pixels
        instead of relative to the parent.
        """
        tmp = self._absolute
        return tmp[0]
    
    @property
    def absTop(self):
        """ Get the y-element of the position, expressed in absolute pixels
        instead of relative to the parent.
        """
        tmp = self._absolute
        return tmp[1]
    
    @property
    def absTopLeft(self):
        """ Get a tuple (absLeft, absTop).
        """
        tmp = self._absolute
        return tmp[0], tmp[1]
    
    @property
    def absRight(self):
        """ Get absLeft+width.
        """
        tmp = self._absolute
        return tmp[0] + tmp[2]
    
    @property
    def absBottom(self):
        """ Get absTop+height.
        """
        tmp = self._absolute
        return tmp[1] + tmp[3]
    
    @property
    def absBottomRight(self):
        """ Get a tuple (right, bottom).
        """
        tmp = self._absolute
        return tmp[0] + tmp[2], tmp[1] + tmp[3]



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" Module motion

Defines the class to create animated data.

"""

import visvis as vv
from visvis import Wobject, Timer
from visvis.core.misc import basestring


class MotionMixin(object):
    """ MotionMixin(axes)
    
    A generic class that represents a wobject that has motion.
    
    This class implements the basic methods to control motion. It uses
    the concept of the motionIndex, which is basically the time parameter.
    Its value specifies the position in the motion sequence with a scalar
    (float) value. Some motion wobject implementations may only support
    integer values though, in which case they should round the motionIndex.
    
    Inheriting classes should implement _GetMotionCount() and
    _SetMotionIndex(index, ii, ww), where ii are four indices, and ww
    are four weights. Together they specify a spline.
    """
    
    def __init__(self):
        
        # Init motion stuff
        self._motionIndex = 0.0
        self._motionIndexDelta = 1.0
        self._motionIsCyclic = True
        self._motionSplineType = 'linear'
        
        # Timer for playing
        self._motionTimer = Timer(self)
        self._motionTimer.Bind(self._MotionTimerCallback)
    
    
    @vv.misc.Property
    def motionIsCyclic():
        """ Get/set whether the motion is cyclic. Default True.
        """
        def fget(self):
            return self._motionIsCyclic
        def fset(self, value):
            self._motionIsCyclic = bool(value)
            self.motionIndex = self.motionIndex
        return locals()
    
    
    @property
    def motionCount(self):
        """ Get the number of temporal instances of this object.
        For objects that do not have a fixed number of elements,
        the returned number may be an arbitrary large number above
        one thousand.
        """
        return self._GetMotionCount()
    
    
    @vv.misc.PropWithDraw
    def motionIndex():
        """ Get/set the motion index; the temporal position in the
        motion. With N the motionCount, if the motion is cyclic, the
        valid range is [0, N>, otherwise [0,N-1]. If the given index
        is not in the valid range, its modulo is taken (cyclic) or
        the index is clipped (not cyclic).
        """
        # This property is where the "reset" basically takes place
        def fget(self):
            return self._motionIndex
        def fset(self, value):
            # Make float
            value = float(value)
            
            # Get N and M (M is index limit)
            N = M = self.motionCount
            if not self.motionIsCyclic:
                M = N-1
            
            # Limit value and set
            if N<2:
                self._motionIndex = 0.0
            else:
                self._motionIndex = self._MotionCorrectIndex(value, M, True)
            
            # Create new deform array
            if N==0 or N==1:
                ii = 0,0,0,0
                ww = 1,0,0,0
            else:
                # Get index and factor
                i1 = int(self._motionIndex)
                t = self._motionIndex - i1
                i0 = self._MotionCorrectIndex(i1-1, M, self._motionIsCyclic)
                i2 = self._MotionCorrectIndex(i1+1, M, self._motionIsCyclic)
                i3 = self._MotionCorrectIndex(i1+2, M, self._motionIsCyclic)
                ii = i0, i1, i2, i3
                # Linear combine
                ww = self._MotionGetCoeff(t, self._motionSplineType)
            
            # Apply
            self._SetMotionIndex(self._motionIndex, ii, ww)
        
        return locals()
    
    
    def _MotionCorrectIndex(self, i, N, cyclic):
        """ _MotionCorrectIndex(i, N, cyclic)
        Correct the given motion index. If cyclic will take the modulo,
        otherwise will clip the index to the valid range.
        """
        if cyclic:
            while i>=N:
                i -= N
            while i<0:
                i += N
        else:
            if i>N:
                i = N
            if i<0:
                i = 0
        return i
    
    
    def _GetMotionCount(self):
        """ _GetMotionCount()
        
        Called to get the number of temporal instances. If there is no limit,
        just return a very large number above thousand. Should be overloaded.
        
        """
        raise NotImplemented()
    
    
    def _SetMotionIndex(self, index, ii, ww):
        """ _SetMotionIndex(index, ii, ww)
        
        Called when the motion index is set. Should be overloaded.
        
        """
        raise NotImplemented()
    
    
    def MotionPlay(self, interval=100, delta=1):
        """ MotionPlay(interval=100, delta=1)
        
        Start playing the motion.
        
        This starts a timer that is triggered every interval miliseconds.
        On every trigger the motionIndex is increased by delta.
        
        """
        self._motionIndexDelta = delta
        self._motionTimer.Start(interval, False)
    
    
    def MotionStop(self):
        """ MotionStop()
        
        Stop playing the motion.
        
        """
        self._motionTimer.Stop()
    
    
    def _MotionTimerCallback(self, event):
        """ _MotionTimerCallback(event)
        Callback for the motion timer.
        """
        if self._destroyed:
            # Auto-stop timer if wobject is destroyed
            self._motionTimer.Stop()
            return
        else:
            # Increase deform index
            self.motionIndex += self._motionIndexDelta

    
    @vv.misc.Property
    def motionSplineType():
        """ Get/set the spline type to interpolate the motion. Can be
        'nearest', 'linear', 'cardinal', 'B-spline', or a float between
        -1 and 1 specifying the tension of the cardinal spline. Note that
        the B-spline is an approximating spline. Default 'linear'. Note that
        an implementation may not support interpolation of the temporal
        instances.
        """
        def fget(self):
            return self._motionSplineType
        def fset(self, value):
            if isinstance(value, (float, int)):
                if value >= -1 and value <= 1:
                    self._motionSplineType = float(value)
                else:
                    raise ValueError('Tension parameter must be between -1 and 1.')
            elif isinstance(value, basestring):
                value = value.lower()
                if value in ['nearest', 'linear', 'cardinal', 'c', 'b-spline', 'b']:
                    self._motionSplineType = value
                else:
                    raise ValueError('Invalid spline type.')
            else:
                raise ValueError('motionSplineType must be scalar or string.')
            self.motionIndex = self.motionIndex
        return locals()
    
    
    def _MotionGetCoeff(self, t, spline_type):
        """ _MotionGetCoeff(t, splineType)
        Get the coefficients to interpolate the motion.
        """
        # Accept tension parameter as spline type
        tension = 0.0
        if isinstance(spline_type, (float, int)):
            tension = float(spline_type)
            spline_type = 'cardinal'
        
        # Get id
        spline_type = spline_type.lower()
        if spline_type in ['c', 'card', 'cardinal', 'catmull–rom']:
            tau = 0.5 * (1.0 - tension)
            w0 = - tau * (   t**3 - 2*t**2 + t )
            w3 =   tau * (   t**3 -   t**2     )
            w1 =           2*t**3 - 3*t**2 + 1  - w3
            w2 = -         2*t**3 + 3*t**2      - w0
        elif spline_type in ['b', 'basis', 'basic']:
            w0 = (1-t)**3                     /6.0
            w1 = ( 3*t**3 - 6*t**2 +       4) /6.0
            w2 = (-3*t**3 + 3*t**2 + 3*t + 1) /6.0
            w3 = (  t)**3                     /6.0
        elif spline_type in ['nn', 'nearest']:
            w0, w3 = 0.0, 0.0
            w1 = float(t<0.5)
            w2 = float(not w1)
        else: # Linear
            w0, w3 = 0.0, 0.0
            w1 = (1.0-t)
            w2 = float(t)
        
        # Return coefficients
        return w0, w1, w2, w3


class MotionSyncer(MotionMixin):
    """ MotionSyncer(*motionWobjects)
    
    Simple class to sync the motion of multiple motion objects.
    
    To sync multiple motion objects, give them to the initializer of this
    class, or register them via the append() method. Then use motionPlay()
    on this class.
    
    """
    def __init__(self, *args):
        MotionMixin.__init__(self)
        self._motionObjects = []
        
        # Append given objects
        for arg in args:
            self.append(arg)
    
    
    def append(self, wobject):
        """  append(wobject)
        
        Append (i.e. register) a motion wobject to be synced.
        
        """
        if not isinstance(wobject, MotionMixin):
            raise ValueError('Given object does not inherit from MotionMixin.')
        else:
            self.remove(wobject)
            self._motionObjects.append(wobject)
    
    
    def remove(self, wobject):
        """ remove(wobject)
        
        Remove (i.e. unregister) a motion wobject to be synced. This
        method does nothing if the given wobject was not in the list
        of wobjects to be synced.
        
        """
        while wobject in self._motionObjects:
            self._motionObjects.remove(wobject)
    
    
    def _GetMotionCount(self):
        """ _GetMotionCount()
        
        Get the smallest number of temporal instances for the
        registered MotionWobjects.
        
        """
        
        # Get count for each child
        counts = [child._GetMotionCount() for child in self._motionObjects]
        
        # Take minumum or return 0
        if counts:
            return min(counts)
        else:
            return 0
    
    
    def Draw(self):
        # To behave a bit like a wobject, so that the OnDraw props work
        for child in self._motionObjects:
            child.Draw()
    
    @property
    def _destroyed(self):
        # To stop if one of the motion objects is destroyed
        for child in self._motionObjects:
            if child._destroyed:
                return True
        else:
            return False
    
    
    def _SetMotionIndex(self, index, ii, ww):
        """ _SetMotionIndex(index, ii, ww)
        
        Sets the motion index for all registered MotionWobjects.
        
        """
        for child in self._motionObjects:
            child._SetMotionIndex(index, ii, ww)


class MotionDataContainer(Wobject, MotionMixin):
    """ MotionDataContainer(parent, interval=100)
    
    The motion data container is a wobject that can contain
    several data, which are displayed as a sequence.
    
    The data are simply stored as the wobject's children, and are
    made visible one at a time.
    
    Example
    -------
    # read image
    ims = [vv.imread('astronaut.png')]

    # make list of images: decrease red channel in subsequent images
    for i in range(9):
        im = ims[i].copy()
        im[:,:,0] = im[:,:,0]*0.9
        ims.append(im)

    # create figure, axes, and data container object
    a = vv.gca()
    m = vv.MotionDataContainer(a)

    # create textures, loading them into opengl memory, and insert into container.
    for im in ims:
        t = vv.imshow(im)
        t.parent = m
    
    """
    
    def __init__(self, parent, interval=100):
        Wobject.__init__(self,parent)
        MotionMixin.__init__(self)
        self.motionSplineType = 'nearest'
        self.MotionPlay(interval)
    
    
    @property
    def timer(self):
        """ Get the timer object used to make the objects visible.
        For backward compatibility.
        """
        return self._motionTimer
    
    
    def _GetMotionCount(self):
        """ _GetMotionCount()
        
        Get the number of temporal instances (i.e. children).
        """
        return len(self.children)
    
    
    def _SetMotionIndex(self, index, ii, ww):
        """ _SetMotionIndex(index, ii, ww)
        
        Make the right child visible.
        
        """
        
        # Get index
        index = int(round(index))
        if index<0:
            index = 0
        if index >= len(self.children):
            index = len(self.children)-1
        
        # set all invisible
        for child in self.children:
            child.visible = False
        
        # except one
        if self.children:
            self.children[index].visible = True
        
        # show it!
        self.Draw()



# -*- coding: utf-8 -*-
# Copyright (C) 2012, Almar Klein
#
# Visvis is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" Module colorWibjects

Implements a few widgets to interactively change color properties of textures.

"""

import OpenGL.GL as gl

import numpy as np
import weakref

from visvis.utils.pypoints import Point, Pointset
from visvis import Colormapable
#
from visvis import Box, DraggableBox
from visvis import Label
from visvis import BaseFigure, Axes
from visvis.core.axises import GetTicks
#
from visvis.wibjects.buttons import RadioButton
from visvis.wibjects.sliders import RangeSlider


class BaseMapableEditor(DraggableBox):
    """ BaseMapableEditor(parent)
    
    Base class for widgets that are used to edit the mapable properties
    of objects: colormap and clim.
    
    """
    
    def __init__(self, parent):
        DraggableBox.__init__(self, parent)
        
        # Objects to map
        self._mapables = []
    
    
    def GetMapables(self):
        """ GetMapables()
        
        Get a list of mappable objects that was given earlier using
        SetMapables. If an axes or figure was given, all eligable
        objects are queried from their children.
        
        """
        
        # Remove dead refs
        self._mapables = [ref for ref in self._mapables if ref() is not None]
        
        # Get the actual mapable objects
        mapables = []
        for arg in self._mapables:
            # Get real ref
            arg = arg()
            if isinstance(arg, (BaseFigure, Axes)):
                mapables.extend(arg.FindObjects(Colormapable))
            elif isinstance(arg, Colormapable):
                mapables.append(arg)
        
        # Done
        return mapables
    
    
    def SetMapables(self, *args):
        """ SetMapables(mapable1, mapable2, mapable3, etc.)
        
        Set the mapables to take into account. A mapable is any
        wibject or wobject that inherits from Colormapable (and can
        be recognized by having a colormap and clim property).
        
        The argument may also be a list or tuple of objects, or an Axes
        or Figure instance, in which case all mapable children are used.
        If args is not given, the parent is used.
        
        """
        
        # Parse input
        if not args:
            args = [self.parent]
        elif len(args)==1 and hasattr(args[0], '__len__'):
            args = args[0]
        if not args:
            return
        
        # Get the weak refs to the actual wobjects
        self._mapables = []
        for arg in args:
            if isinstance(arg, (BaseFigure,Axes)) or isinstance(arg, Colormapable):
                self._mapables.append( weakref.ref(arg) )
            else:
                print('Warning object is not mappable: %s' % str(arg))


class ClimEditor(BaseMapableEditor):
    """ CLimEditor(parent, *args)
    
    A wibject to edit the clim property of textures (thereby setting
    window-width and window-level).
    
    During initialization, SetMapables(*args) is called. The easiest way
    to use this wibject is to attach it to an axes or figure instance.
    The wibject then controls the colormaps of all mapable objects in them.
    
    """
    
    def __init__(self, parent, *args):
        BaseMapableEditor.__init__(self, parent)
        
        # init size
        self.position.w = 300
        self.position.h = 50
        
        # Init mappables
        self._mapables = []
        
        # Create slider widget
        self._slider = RangeSlider(self)
        self._slider.position = 15,5,-30,-30 # 80 = 55+25
        self._slider.showTicks = True
        
        # Bind events
        self._slider.eventSliderChanged.Bind(self._UpdateFull)
        
        # Set mappables
        self.SetMapables(*args)
        self._InitFromMapables()
    
    
    @property
    def slider(self):
        """ Get the slider instance of this tool.
        """
        return self._slider
    
    def _UpdateFull(self, event):
        for mappable in self.GetMapables():
            ra = self._slider.range
            if hasattr(mappable, 'SetClim'):
                mappable.SetClim(ra.min, ra.max)
            else:
                mappable.clim = ra.min, ra.max
    
    
    def _InitFromMapables(self):
        """ _InitFromMapables()
        
        The clim of the last mapable obtained by GetMapables is used
        to initialize the editor (if possible).
        
        """
        
        # Get the last one and obtain its limits
        mappables = self.GetMapables()
        if mappables:
            # Update widget
            data = mappables[-1]._texture1._dataRef
            self._slider.fullRange = data.min(), data.max()
            self._slider.range = mappables[-1].clim


class ColormapEditor(BaseMapableEditor):
    """ ColormapEditor(parent *args)
    
    A wibject to edit colormaps.
    
    During initialization, SetMapables(*args) is called. The easiest way
    to use this wibject is to attach it to an axes or figure instance.
    The wibject then controls the colormaps of all mapable objects in them.
    
    """
    
    def __init__(self, parent, *args):
        BaseMapableEditor.__init__(self, parent)
        
        # init size
        self.position.w = 300
        self.position.h = 80
        
        # Init mappables
        self._mapables = []
        
        # Create node widget
        self._nodeWidget = CM_NodeWidget(self)
        self._nodeWidget.position = 55,35,-80,-40 # 80 = 55+25
        
        # Create buttons
        self._rBut = RadioButton(self, 'R'); self._rBut.position =  5,35, 12,14
        self._gBut = RadioButton(self, 'G'); self._gBut.position = 20,35, 12,14
        self._bBut = RadioButton(self, 'B'); self._bBut.position = 35,35, 12,14
        self._aBut = RadioButton(self, 'A'); self._aBut.position =  5,52, 12,14
        self._aBut.state = True
        
        # Create colorbar wibject
        cb = Colorbar(self)
        pos = self._nodeWidget.position
        cb.position = pos.x, 5, pos.w, 10
        
        # Bind events
        self._aBut.eventStateChanged.Bind(self._OnChannelSelect)
        
        # Bind events to THIS wibject, and process them in the nodeWidget
        # This is to allow the user to click just outside the wibject also.
        self.eventMouseDown.Bind(self._OnDown)
        self.eventMouseUp.Bind(self._OnUp)
        self.eventDoubleClick.Bind(self._OnDoubleClick)
        
        # Set mappables
        self.SetMapables(*args)
        self._InitFromMapables()
    
    
    def _OnDown(self, event):
        """ Pass event on """
        # Get event of nodeWibject and its position
        event2 = self._nodeWidget.eventMouseDown
        pos = self._nodeWidget.position
        # Calc location and limit
#         event2.Set(event.x-pos.left, event.y-pos.top, event.button)
        event2.Set(event.absx, event.absy, event.button)
        # Fire!
        if event2.x > -10 and event2.x < pos.width + 10:
            if event2.y > -10 and event2.y < pos.height + 10:
                event2.Fire()
                return True # Prevent dragging
    
    def _OnUp(self, event):
        """ Pass event on """
        # Get event of nodwWibject and its position
        event2 = self._nodeWidget.eventMouseUp
        event2.Set(event.absx, event.absy, event.button)
        # Fire!
        event2.Fire()
    
    
    def _OnDoubleClick(self, event):
        """ Pass event on """
        # get event of nodwWibject and its position
        event2 = self._nodeWidget.eventDoubleClick
        
        # Calc location and limit
#         pos = self._nodeWidget.position
#         event2.Set(event.x-pos.left, event.y-pos.top, event.button)
        event2.Set(event.absx, event.absy, event.button)
        # Fire!
        event2.Fire()
    
    
    def _OnChannelSelect(self, event):
        
        # Apply
        if self._rBut.state:
            self._nodeWidget._nodes = self._nodeWidget._allNodes[0]
            self._nodeWidget._line = self._nodeWidget._allLines[0]
        elif self._gBut.state:
            self._nodeWidget._nodes = self._nodeWidget._allNodes[1]
            self._nodeWidget._line = self._nodeWidget._allLines[1]
        elif self._bBut.state:
            self._nodeWidget._nodes = self._nodeWidget._allNodes[2]
            self._nodeWidget._line = self._nodeWidget._allLines[2]
        elif self._aBut.state:
            self._nodeWidget._nodes = self._nodeWidget._allNodes[3]
            self._nodeWidget._line = self._nodeWidget._allLines[3]
        
        # Show
        self._nodeWidget._OnUp()
    
    
    def _InitFromMapables(self):
        """ _InitFromMapables()
        
        The colormap of the last mapable obtained by GetMapables is used
        to initialize the editor (if possible).
        
        """
        
        # Get the last mappable and obtain its colormap
        mappables = self.GetMapables()
        if not mappables:
            return
        cmap = mappables[-1]._GetColormap()
        if not cmap:
            return
        
        # Get node data
        nodeData = cmap
        
        # Can we make nodes of it?
        if isinstance(nodeData, list):
            
            # Obtain position
            tt = np.linspace(0,1,len(nodeData))
            
            # Clear nodes
            for i in range(4):
                self._nodeWidget._allNodes[i].clear()
                
            # Create nodes
            for t, node in zip(tt,nodeData):
                for i in range(4):
                    nodes = self._nodeWidget._allNodes[i]
                    if i< len(node):
                        nodes.append(t,1-node[i])
                    else:
                        nodes.append(t,0)
        
        elif isinstance(nodeData, dict):
            # Allow several color names
            for key in list(nodeData.keys()):
                if key.lower() in ['r', 'red']:
                    nodeData['r'] = nodeData[key]
                elif key.lower() in ['g', 'green']:
                    nodeData['g'] = nodeData[key]
                if key.lower() in ['b', 'blue']:
                    nodeData['b'] = nodeData[key]
                if key.lower() in ['a', 'alpha']:
                    nodeData['a'] = nodeData[key]
            
            # Create nodes
            for i in range(4):
                key = 'rgba'[i]
                if key in nodeData:
                    nodes = self._nodeWidget._allNodes[i]
                    nodes.clear()
                    for t, val in nodeData[key]:
                        nodes.append(t,1-val)
        
        # Update
        self._nodeWidget._UpdateFull()

    
class CM_NodeWidget(Box):
    """ CM_NodeWidget(parent)
    
    Class to modify node positions using the mouse.
    
    """
    
    def __init__(self, parent):
        Box.__init__(self, parent)
        
        # Set color
        self.bgcolor = 'w'
        
        # How soon the mouse snaps to a node
        self._snapWidth = 5
        
        # The currently selected node (or None if nothing is selected)
        self._selectedNode = None
        
        # The nodes and lines of R,G,B and A
        self._allNodes = [Pointset(2) for i in range(4)]
        self._allLines = [Pointset(2) for i in range(4)]
        
        # Init nodes
        for nodes in self._allNodes:
            nodes.append(0,1)
            nodes.append(1,0)
        self._allNodes[3][0,1] = 0 # alpha is completele ones by default
        
        # Init lines
        for line in self._allLines:
            for i in np.linspace(0,1,256):
                line.append(i,1) # the actual map will be 1-line
        
        # Make lines correct now
        for nodes, line in zip(self._allNodes, self._allLines):
            self._NodesToLine(nodes, line)
        
        # The node and line currently in control
        self._nodes = self._allNodes[3]
        self._line = self._allLines[3]
        
        # Bind events
        self.eventMotion.Bind(self._OnMotion)
        self.eventMouseUp.Bind(self._OnUp)
        self.eventMouseDown.Bind(self._OnDown)
        self.eventDoubleClick.Bind(self._OnDoubleClick)
    
    
    def _OnDown(self, event):
        
        # calculate distance of mouse to all points
        p = Point(event.x, event.y)
        dists = p.distance( self._nodes * Point(self.position.size) )
        
        # is one close enough?
        i = -1
        if self._nodes: # or .min() will not work
            tmp = dists.min()
            if tmp <= self._snapWidth:
                i, = np.where(dists==tmp)
        
        # if so, store that point
        if i>= 0:
            self._selectedNode = i
        else:
            self._selectedNode = None
    
    
    def _OnUp(self, event=None):
        if self._selectedNode is None:
            return
        
        # clean up
        self._selectedNode =  None
        
        # Calculate the lines one last time
        self._NodesToLine(self._nodes, self._line, True)
        
        # Draw (not fast)
        self.Draw()
    
    
    def _OnMotion(self, event):
        
        # should we proceed?
        if self._selectedNode is None:
            return
        if self._selectedNode >= len(self._nodes):
            self._OnUp()
            return
        
        # calculate and limit new position
        x, y = event.x, event.y
        pos = Point(x,y) / Point(self.position.size)
        if pos.x<0: pos.x=0
        if pos.y<0: pos.y=0
        if pos.x>1: pos.x = 1
        if pos.y>1: pos.y = 1
        
        # Change node's position
        self._nodes[self._selectedNode,0] = pos.x
        self._nodes[self._selectedNode,1] = pos.y
        
        # Apply to line
        self._NodesToLine(self._nodes, self._line)
        
        # Draw (fast)
        self.Draw(True)
    
    
    def _OnDoubleClick(self, event):
        
        # use OnDown method to test if and which node was clicked
        self._OnDown(event)
        
        if self._selectedNode is None:
            # create new node?
            pos = self.position
            x, y = event.x, event.y
            if x>-5 and x<pos.width+5 and y>-5 and y<pos.height-5:
                pos = Point(event.x, event.y) / Point(self.position.size)
                self._nodes.append( pos )
        else:
            # remove point
            self._nodes.pop(self._selectedNode)
        # Update
        self._UpdateFull()
    
    
    def _UpdateFull(self):
        """ Update all lines and draw. """
        
        # Create the lines
        for nodes, line in zip(self._allNodes, self._allLines):
            self._NodesToLine(nodes, line)
        
        # Draw
        self.Draw(True)
    
    
    def _NodesToLine(self, nodes, line, update=False):
        """ Convert nodes to a full 256 element line.
        """
        
        # sort nodes
        nn = [n for n in nodes]
        nn.sort(key=lambda n:n.x)
        nodes = Pointset(2)
        for n in nn:
            nodes.append(n)
        
        # interpolate
        xx = np.linspace(0,1,256)
        if nodes:
            yy = np.interp(xx, nodes[:,0], nodes[:,1])
            if np.isnan(yy[-1]): # happens when last two nodes on same pos
                yy[-1] = yy[-2]
            line[:,1] = yy
        else:
            line[:,1] = np.zeros((256,),dtype=np.float32) # no nodes
        
        if update:
            # Create colormap
            map = {}
            for i in range(4):
                nn = self._allNodes[i]
                tmp = []
                for ii in range(len(nn)):
                    tmp.append((nn[ii,0], 1-nn[ii,1]))
                if tmp:
                    key = 'rgba'[i]
                    map[key] = sorted(tmp, key=lambda x:x[0])
            
            # Apply colormap to all registered objects
            if self.parent:
                for ob in self.parent.GetMapables():
                    ob._SetColormap(map)
                    ob.Draw()
    
    
    def OnDraw(self):
        Box.OnDraw(self)
        
        # Create color dict
        colors = {  self._allLines[0]:(1,0,0), self._allLines[1]:(0,1,0),
                    self._allLines[2]:(0,0,1), self._allLines[3]:(0,0,0) }

        # prepare scaling
        gl.glPushMatrix()
        w,h = self.position.size
        gl.glScale(w, h, 1)
        
        
        # prepare
        gl.glPointSize(7)
        gl.glLineWidth(1)
        gl.glEnable(gl.GL_POINT_SMOOTH) # round points
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        
        # Draw lines
        for line in self._allLines:
            if len(line) and line is not self._line:
                gl.glColor( *colors[line] )
                gl.glVertexPointerf(line.data)
                gl.glDrawArrays(gl.GL_LINE_STRIP, 0, len(line))
        
        # Draw the line under control (using a thicker line)
        gl.glColor( *colors[self._line] )
        gl.glLineWidth(2)
        gl.glVertexPointerf(self._line.data)
        gl.glDrawArrays(gl.GL_LINE_STRIP, 0, len(self._line))
        
        # draw nodes
        gl.glColor( *colors[self._line] )
        gl.glVertexPointerf(self._nodes.data)
        gl.glDrawArrays(gl.GL_POINTS, 0, len(self._nodes))
        
        # clean up
        gl.glPopMatrix()
        gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
        gl.glDisable(gl.GL_POINT_SMOOTH)


class Colorbar(Box):
    """ Colorbar(parent)
    
    This wibject displays the colormap applied to a certain mapable.
    It should be attached to an Axes or Figure. The displayed colormap
    is that of the last added (or last added in the last axes) mapable
    object.
    
    A Colorbar can be created with the function vv.colorbar().
    
    """
    
    def __init__(self, parent):
        Box.__init__(self, parent)
        
        self._label = Label(self, '')
        self._label.bgcolor = ''
        self.eventPosition.Bind(self._OnPositionChange)
        
        # Pool of labels for tickmarks, to reuse them
        self._wobjects = [] # So we can have Text objects
        self._labelPool = {}
        
        # If attached to axes, correct that axes' size
        if isinstance(parent, Axes):
            
            # Init position
            x = parent.position.width + 5
            self.position = x, 0.0, 30, 1.0
            
            # Keep informed of axes movment
            self.parent.eventPosition.Bind(self._OnAxesPositionChange)
            
            # Correct axes' position
            self.parent.position.Correct(dw=-100) # 30 + 70
    
    
    def _OnPositionChange(self, event):
        """ Adjust the position of the label.
        """
        self._label.halign = 'center'
        self._label.valign = 'center'

        w,h = self.position.size
        if w > h:
            self._label.textAngle = 0
            self._label.position.w = 100
            self._label.position.h = self._label.fontSize
            self._label.position.x = (w - 100)/2.
            self._label.position.y = h + 20
        else:
            self._label.textAngle = 90
            self._label.position.w = self._label.fontSize
            self._label.position.h = 100
            self._label.position.x = w + 40
            self._label.position.y = (h - 100)/2.
    
    def _OnAxesPositionChange(self, event):
        axes = event.owner
        if axes:
            self.position.x = axes.position.width+5
    
    @property
    def label(self):
        """ Get the label instance for this colorbar.
        """
        return self._label
    
    
    def SetLabel(self, value):
        """ SetLabel(value)
        
        A convenience function to set the label text.
        
        """
        self._label.text = value
    
    
    def OnDraw(self):
        
        # Get colormaps that apply
        par = self.parent
        if par is None:
            return
        elif isinstance(par, (BaseFigure, Axes)):
            mapables = par.FindObjects(Colormapable)
        elif isinstance(par, ColormapEditor):
            mapables = par.GetMapables()
        elif isinstance(par, ClimEditor):
            mapables = par.GetMapables()
        else:
            mapables = []
        
        # Get the last one
        mapable = None
        if mapables:
            mapable = mapables[-1]
        
        
        # get dimensions
        w,h = self.position.size
        
        # Calc map direction
        if w > h:
            texCords = [0,0,1,1]
        else:
            texCords = [1,0,0,1]
        
        
        # draw plane
        if mapable:
            # Use it's colormap texture
            mapable._EnableColormap()
            # Disable alpha channel (by not blending)
            gl.glDisable(gl.GL_BLEND)
            gl.glColor(1.0, 1.0, 1.0, 1.0)
            # Draw quads
            gl.glBegin(gl.GL_QUADS)
            gl.glTexCoord1f(texCords[0]); gl.glVertex2f(0,0)
            gl.glTexCoord1f(texCords[1]); gl.glVertex2f(0,h)
            gl.glTexCoord1f(texCords[2]); gl.glVertex2f(w,h)
            gl.glTexCoord1f(texCords[3]); gl.glVertex2f(w,0)
            gl.glEnd()
            
            # Clean up
            gl.glEnable(gl.GL_BLEND)
            gl.glFlush()
            mapable._DisableColormap()
        
        # prepare
        gl.glDisable(gl.GL_LINE_SMOOTH)
        
        # draw edges
        if self.edgeWidth and self.edgeColor:
            clr = self.edgeColor
            gl.glColor(clr[0], clr[1], clr[2], 1.0)
            gl.glLineWidth(self.edgeWidth)
            #
            gl.glBegin(gl.GL_LINE_LOOP)
            gl.glVertex2f(0,0)
            gl.glVertex2f(0,h)
            gl.glVertex2f(w,h)
            gl.glVertex2f(w,0)
            gl.glEnd()
        
        if hasattr(mapable, 'clim'):
            # Draw ticks
            if w>h:
                p0 = Point(0, h)
                p1 = Point(w, h)
                delta = Point(0,3)
                halign, valign = 0, 0
                xoffset, yoffset = -8, -2
            else:
                p0 = Point(w, h)
                p1 = Point(w, 0)
                delta = Point(3,0)
                halign, valign = -1, 0
                xoffset, yoffset = 5, -8
            
            # Get tickmarks
            ticks, ticksPos, ticksText = GetTicks(p0, p1, mapable.clim)
            
            newLabelPool = {}
            linePieces = Pointset(2)
            for tick, pos, text in zip(ticks, ticksPos, ticksText):
                pos2 = pos + delta
                
                # Add line piece
                linePieces.append(pos); linePieces.append(pos2)
                
                # Create or reuse label
                if tick in self._labelPool:
                    label = self._labelPool.pop(tick)
                else:
                    label = Label(self, ' '+text+' ')
                    label.bgcolor = ''
                
                # Position label and set text alignment
                newLabelPool[tick] = label
                label.halign, label.valign = halign, valign
                label.position.x = pos2.x + xoffset
                label.position.w = 16
                label.position.y = pos2.y + yoffset
            
            # Clean up label pool
            for label in list(self._labelPool.values()):
                label.Destroy()
            self._labelPool = newLabelPool
            
            # Draw line pieces
            # prepare
            gl.glLineWidth(1)
            gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
            gl.glVertexPointerf(linePieces.data)
            gl.glDrawArrays(gl.GL_LINES, 0, len(linePieces))
            gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
        # clean up
        gl.glEnable(gl.GL_LINE_SMOOTH)



#!/usr/bin/env python
import visvis as vv
import OpenGL.GL as gl


class CustomWobject(vv.Wobject):
    """ Example Custom wobject.
    This example is not optimal, it is just to illustrate how Wobject
    can be subclassed.
    """
    
    def __init__(self, parent):
        vv.Wobject.__init__(self, parent)
    
    def _drawTraingles(self, how, color):
        a = 0, 0, 0
        b = 1, 0, 0
        c = 1, 1, 0
        d = 1, 1, 1
        
        gl.glColor(*color)
        gl.glBegin(how)
        gl.glVertex(*a);  gl.glVertex(*b); gl.glVertex(*c)
        gl.glVertex(*a);  gl.glVertex(*c); gl.glVertex(*d)
        gl.glVertex(*b);  gl.glVertex(*c); gl.glVertex(*d)
        gl.glVertex(*a);  gl.glVertex(*d); gl.glVertex(*b)
        gl.glEnd()
    
    def _GetLimits(self):
        """ Tell the axes how big this object is.
        """
        # Get limits
        x1, x2 = 0, 1
        y1, y2 = 0, 1
        z1, z2 = 0, 1
        # Return
        return vv.Wobject._GetLimits(self, x1, x2, y1, y2, z1, z2)
    
    def OnDraw(self):
        """ To draw the object.
        """
        self._drawTraingles(gl.GL_TRIANGLES, (0.2, 0.8, 0.4))
        gl.glLineWidth(3)
        self._drawTraingles(gl.GL_LINE_LOOP, (0,0,0))
    
    def OnDrawShape(self, clr):
        """ To draw the shape of the object.
        Only necessary if you want to be able to "pick" this object
        """
        self._drawTraingles(gl.GL_TRIANGLES, clr)
    
    def OnDrawScreen(self):
        """ If the object also needs to draw in screen coordinates.
        Text needs this for instance.
        """
        pass
    
    def OnDestroyGl(self):
        """ To clean up any OpenGl resources such as textures or shaders.
        """
        pass
    
    def OnDestroy(self):
        """ To clean up any other resources.
        """
        pass
    

# Create an instance of this class and set axes limits
a = vv.cla()
c = CustomWobject(a)
a.SetLimits()

# Enter main loop
app = vv.use() # let visvis chose a backend for me
app.Run()
